use std::{collections::HashSet, io::ErrorKind, path::Path};

use axum::{
    extract::{BodyStream, Extension},
    http::StatusCode,
    response::IntoResponse,
};
use flate2::read::GzDecoder;
use futures_util::TryStreamExt;
use tar::Archive;
use tokio_util::io::StreamReader;

use crate::STATIC_DIR_NAME;

pub async fn post_deploy(
    Extension(data_dir): Extension<String>,
    res: BodyStream,
) -> impl IntoResponse
{
    // This converts a stream into something that implements AsyncRead, which we can then use to
    // asynchronously write the file to disk
    let mut read =
        StreamReader::new(res.map_err(|axum_err| std::io::Error::new(ErrorKind::Other, axum_err)));
    let uuid = uuid::Uuid::new_v4();
    let file_path = Path::new(&data_dir).join(uuid.as_hyphenated().to_string());
    let mut file = tokio::fs::File::create(&file_path).await.unwrap();
    tokio::io::copy(&mut read, &mut file).await;

    // Extract the contents of the tarball synchronously
    match tokio::task::spawn_blocking(move || {
        let file = match std::fs::File::open(file_path) {
            Ok(v) => v,
            Err(_) => return StatusCode::INTERNAL_SERVER_ERROR,
        };
        let tar = GzDecoder::new(file);
        let mut archive = Archive::new(tar);

        let mut paths = HashSet::new();

        let entries = match archive.entries() {
            Ok(e) => e,
            Err(_) => return StatusCode::INTERNAL_SERVER_ERROR,
        };

        // Extract each entry into the output directory
        let static_dir = Path::new(&data_dir).join(STATIC_DIR_NAME);
        for entry_res in entries {
            if let Ok(mut entry) = entry_res {
                if let Err(_) = entry.unpack_in(&static_dir) {
                    return StatusCode::INTERNAL_SERVER_ERROR;
                }

                if let Ok(path) = entry.path() {
                    paths.insert(path.into_owned());
                }
            } else {
                return StatusCode::INTERNAL_SERVER_ERROR;
            }
        }

        // Remove any old files that weren't present in new archive
        let mut items = vec![];

        // Start by populating the vec with the initial files
        let iter = match static_dir.read_dir() {
            Ok(v) => v,
            Err(_) => return StatusCode::INTERNAL_SERVER_ERROR,
        };
        iter.filter_map(|r| r.ok())
            .for_each(|e| items.push(e.path()));

        // As long as there are still items in the vec, we keep going
        while items.len() > 0 {
            let item = items.pop().unwrap();
            tracing::debug!("{:?}", item);

            if !paths.contains(item.strip_prefix(&static_dir).unwrap()) {
                if item.is_dir() {
                    std::fs::remove_dir_all(item);
                } else {
                    std::fs::remove_file(item);
                }
            } else if let Ok(iter) = item.read_dir() {
                iter.filter_map(|r| r.ok())
                    .for_each(|e| items.push(e.path()));
            }
        }

        StatusCode::OK
    })
    .await
    {
        Ok(s) => s,
        Err(_) => StatusCode::INTERNAL_SERVER_ERROR,
    }
}
